/*
 * delta_times.c
 *
 * author: Roberto Piassi Passos Bodo
 *  email: rppbodo@gmail.com
 */

#include "delta_times.h"

/*!
 * This function allocates memory for a local delta time structure and makes a
 * temporary copy of the delta time table. It returns 0 in case of success
 * and returns 1 otherwise.
 */
int init_delta_times() {
	int i, k;
	Link temp = NULL;

	delta_times = (double**) malloc(n_tracks * sizeof(double*));
	for (i = 0; i < n_tracks; i++) {
		delta_times[i] = (double*) malloc(events[i] * sizeof(double));
	}

	for (i = 0; i < n_tracks; i++) {
		temp = table[i];
		for (k = 0; k < events[i]; k++) {
			delta_times[i][k] = temp->delta_time;
			temp = temp->next;
		}

		if (k != events[i]) {
			return 1;
		}
	}

	return 0;
}

/*!
 * This function converts the orignal relative time unit (ticks) to the chosen
 * absolute time unit (miliseconds). It returns 0 in case of success and returns
 * 1 otherwise.
 */
int convert_delta_times_() {
	int i, j;
	int first, dt, last, aux;
	Link_t temp = NULL;
	Link link = NULL;

	for (i = 0; i < n_tracks; i++) {
		link = table[i];

		if (link == NULL) {
			continue;
		}

		temp = head_t;
		dt = 0;
		first = 1;

		last = temp->tempo;
		temp = temp->next;

		if (temp != NULL && temp->delta_time < link->delta_time) {
			while (temp != NULL && temp->delta_time < link->delta_time) {
				if (first) {
					dt += temp->delta_time / ((double) ticks_per_beat
							/ (double) last * 1000);
					first = 0;
				} else {
					dt += (temp->delta_time - aux) / ((double) ticks_per_beat
							/ (double) last * 1000);
				}
				aux = temp->delta_time;
				last = temp->tempo;
				temp = temp->next;
			}

			dt += (link->delta_time - aux) / ((double) ticks_per_beat
					/ (double) last * 1000);
		} else {
			dt = link->delta_time / ((double) ticks_per_beat / (double) last
					* 1000);
		}

		delta_times[i][0] = dt;

		j = 1;
		while (link != NULL && link->next != NULL) {
			temp = head_t;

			first = 1;
			dt = 0;

			/* FIXME this part of the function sweeps the entire list all the time! */
			while (temp != NULL && temp->delta_time < link->delta_time) {
				last = temp->tempo;
				temp = temp->next;
			}

			if (temp != NULL && temp->delta_time < link->next->delta_time) {
				while (temp != NULL && temp->delta_time
						< link->next->delta_time) {
					if (first) {
						dt += (temp->delta_time - link->delta_time)
								/ ((double) ticks_per_beat / (double) last
										* 1000);
						first = 0;
					} else {
						dt += (temp->delta_time - aux)
								/ ((double) ticks_per_beat / (double) last
										* 1000);
					}
					aux = temp->delta_time;
					last = temp->tempo;
					temp = temp->next;
				}

				dt += (link->next->delta_time - aux) / ((double) ticks_per_beat
						/ (double) last * 1000);
			} else {
				dt = (link->next->delta_time - link->delta_time)
						/ ((double) ticks_per_beat / (double) last * 1000);
			}

			delta_times[i][j] = dt;

			j++;

			link = link->next;
		}

		if (j != events[i]) {
			return 1;
		}
	}

	for (i = 0; i < n_tracks; i++) {
		for (j = 1; j < events[i]; j++) {
			delta_times[i][j] += delta_times[i][j - 1];
		}
	}

	return 0;
}

/*!
 * This function copies back the delta times from the local structure to the
 * original delta time table. It returns 0 in case of success and returns 1
 * otherwise.
 */
int restore_delta_times() {
	int i, k;
	Link temp;

	for (i = 0; i < n_tracks; i++) {
		temp = table[i];
		k = 0;
		while (temp != NULL) {
			temp->delta_time = delta_times[i][k++];
			temp = temp->next;
		}
		if (k != events[i]) {
			return 1;
		}
	}
	return 0;
}

/*!
 * This function deallocates the memory of the local delta time structure.
 */
void finish_delta_times() {
	int i;

	for (i = 0; i < n_tracks; i++) {
		free(delta_times[i]);
	}
	free(delta_times);
}

/*!
 * This function is called by the main function and it handles the other
 * functions above. It takes care of the return values as well. It returns
 * 0 in case of success and returns 1 otherwise.
 */
int convert_delta_times() {
	if (init_delta_times()) {
		fprintf(stderr, "error: initializing delta times...");
		return 1;
	}

	if (convert_delta_times_()) {
		fprintf(stderr, "error: converting delta times...");
		return 1;
	}

	if (restore_delta_times()) {
		fprintf(stderr, "error: restoring delta times...");
		return 1;
	}

	finish_delta_times();

	return 0;
}
