/*
 * init_functions.c
 *
 * author: Roberto Piassi Passos Bodo
 *  email: rppbodo@gmail.com
 */

#include "init_functions.h"

/*!
 * This function receives the filename string and performs a simple check
 * to find out the file's extension. It returns 1 in case of a '.mid' file,
 * 2 in case of a '.csv' file and 0 otherwise.
 */

int extension(char* filename) {
	int i;
	int end = strlen(filename);

	i = end;
	while (filename[i] != '.') {
		i--;
	}

	if (filename[i + 1] == 'm') {
		if (filename[i + 2] == 'i') {
			if (filename[i + 3] == 'd') {
				return 1;
			}
		}
	}

	if (filename[i + 1] == 'c') {
		if (filename[i + 2] == 's') {
			if (filename[i + 3] == 'v') {
				return 2;
			}
		}
	}

	return 0;
}

/*!
 * This function shows the general usage information.
 */
void help(int argc, char** argv) {
	printf("\n");
	printf("usage: %s [input] [options]\n", argv[0]);
	printf("\n");
	printf("INPUT: '.mid' or '.csv'\n");
	printf("\n");
	printf("OPTIONS:\n");
	printf("   -h    to show general usage information\n");
	printf("   -p    to play samples of all tracks\n");
	printf("   -v    to enable verbose mode which shows all the errors committed by the user\n");
	printf("\n");
}

/*!
 * This function checks the arguments that the user has chosen. First it assigns
 * the input filename to the respective variable and then it sets 1 to the 'player'
 * variable if one of the arguments is '-p' and sets 1 to the 'verbose' variable
 * if one of the arguments is '-d'. It returns 0 in case of success and returns 1
 * otherwise.
 */
int handle_arguments(int argc, char** argv) {
	int i, j;

	verbose = player = 0;
	input_file = NULL;

	if (argc < 2) {
		fprintf(stderr, "error: invalid number of arguments...\n");
		help(argc, argv);
		return 1;
	}

	for (i = 1; i < argc; i++) {
		if (argv[i][0] == '-') {
			for (j = 1; j < strlen(argv[i]); j++) {
				if (argv[i][j] == 'h') {
					help(argc, argv);
					return 1;
				} else if (argv[i][j] == 'v') {
					verbose = 1;
				} else if (argv[i][j] == 'p') {
					player = 1;
				} else {
					fprintf(stderr, "error: invalid arguments...\n");
					help(argc, argv);
					return 1;
				}
			}
		} else {
			if (input_file == NULL) {
				int length = strlen(argv[i]);
				input_file = (char*) malloc(length * sizeof(char));
				strcpy(input_file, argv[i]);
				input_file[length] = '\0';
			} else {
				fprintf(stderr, "error: invalid arguments...\n");
				help(argc, argv);
				return 1;
			}
		}
	}

	if (input_file == NULL) {
		fprintf(stderr, "error: missing filename...\n");
		return 1;
	}

	return 0;
}

/*!
 * This function scans the '/dev' directory and prints all the available midi
 * devices. Then opens the input and output devices chosen by the user. It
 * returns 0 in case of success and returns 1 otherwise.
 */
int set_midi_devices() {
	struct dirent** namelist;
	char input_device[11];
	char output_device[11];
	int i, n, midi_devices;

	n = scandir("/dev", &namelist, 0, alphasort);

	printf("list of available midi devices:\n");
	midi_devices = 0;
	for (i = 0; i < n; i++) {
		if (strncmp(namelist[i]->d_name, "midi", 4) == 0) {
			printf("(%d) /dev/%s\n", (midi_devices + 1), namelist[i]->d_name);
			midi_devices++;
		}
		free(namelist[i]);
	}
	free(namelist);

	if (midi_devices < 2) {
		fprintf(stderr, "\nerror: you don't have enough midi devices...\n");
		return 1;
	}

	printf("\nplease, type the input device...\n");
	scanf("%s", input_device);
	input_device[10] = '\0';

	input_device_fd = open(input_device, O_RDONLY, 0);

	if (input_device_fd < 0) {
		fprintf(stderr, "error: cannot open %s...\n", input_device);
		return 1;
	}

	printf("\nplease, type the output device...\n");
	scanf("%s", output_device);
	output_device[10] = '\0';

	output_device_fd = open(output_device, O_WRONLY, 0);

	if (output_device_fd < 0) {
		fprintf(stderr, "error: cannot open %s...\n", output_device);
		return 1;
	}

	return 0;
}

/*!
 * This function opens the chosen input file and calls the respective function
 * for each valid extension to interpret its bytes. It returns 0 in case of
 * success and returns 1 otherwise.
 */
int load_input_file() {
	FILE* input;
	int ext;

	input = fopen(input_file, "r");

	if (input == NULL) {
		fprintf(stderr, "error: file not found...\n");
		return 1;
	}

	printf("\"%s\"\n", input_file);

	ext = extension(input_file);
	if (ext == 1) {
		if (load_midi(input)) {
			return 1;
		}
	} else if (ext == 2) {
		if (load_csv(input)) {
			return 1;
		}
	} else {
		fprintf(stderr, "error: invalid file extension...\n");
		return 1;
	}

	fclose(input);

	return 0;
}
