/*
 * midi.c
 *
 * author: Roberto Piassi Passos Bodo
 *  email: rppbodo@gmail.com
 */

#include "midi.h"

/*!
 * This function receives n bytes and convert them to an integer.
 */
int bytes_to_decimal(unsigned char* bytes, int n) {
	int i, d = 0;

	for (i = 0; i < n; i++) {
		d = (d << 8);
		d = (d | bytes[i]);
	}

	return d;
}

/*!
 * This function receives k bytes and convert them to an integer, according to
 * the sintax of the Variable Length Values presented in the MIDI Specification.
 */
int variable_time(unsigned char *bytes, int *k) {
	int t = (0x7f & bytes[(*k)]);

	while ((0x80 & bytes[(*k)++]) >> 7) {
		t = (t << 7);
		t = (t | (0x7f & bytes[(*k)]));
	}

	return t;
}

/*!
 * This function interprets the bytes of 'system exclusive' events.
 */
void system_exclusive(unsigned char *bytes, int *k) {
	int i, length = variable_time(bytes, k);

	for (i = 0; i < length; i++) {
		(*k)++;
	}
}

/*!
 * This function interprets the bytes of 'meta' events.
 */
void meta(unsigned char *bytes, int *k, int delta_time) {
	unsigned char type = bytes[(*k)++];
	int length = variable_time(bytes, k);

	switch (type) {
	case (0x51): {
		unsigned char tempo[3];

		tempo[0] = bytes[(*k)++];
		tempo[1] = bytes[(*k)++];
		tempo[2] = bytes[(*k)++];

		insert_tempo(delta_time, bytes_to_decimal(tempo, 3));
		break;
	}
	default: {
		int i;
		for (i = 0; i < length; i++) {
			(*k)++;
		}
		break;
	}
	}
}

/*!
 * This function interprets the bytes of 'midi' events.
 */
void midi(unsigned char *bytes, int *k, int type, int track, int delta_time) {
	unsigned char event_type_value = (type & 0xf0) >> 4;
	unsigned char midi_channel = (type & 0xf);
	unsigned char b[3];

	switch (event_type_value) {
	/* note off event */
	case (0x8): {
		b[0] = 128 + midi_channel;
		b[1] = bytes[(*k)++];
		b[2] = bytes[(*k)++];

		insert_table(track, delta_time, b);
		break;
	}
	/* note on event */
	case (0x9): {
		b[0] = 144 + midi_channel;
		b[1] = bytes[(*k)++];
		b[2] = bytes[(*k)++];

		insert_table(track, delta_time, b);
		break;
	}
	/* note aftertouch event */
	case (0xa): {
		(*k) += 2;
		break;
	}
	/* controller event */
	case (0xb): {
		b[0] = 176 + midi_channel;
		b[1] = bytes[(*k)++];
		b[2] = bytes[(*k)++];

		insert_table(track, delta_time, b);
		break;
	}
	/* program change event */
	case (0xc): {
		int instrument = bytes[(*k)++];
		b[0] = 192 + midi_channel;
		b[1] = instrument;

		printf("track #%d has: %s\n", track, instruments[instrument]);

		write(output_device_fd, b, 2);

		break;
	}
	/* channel aftertouch event */
	case (0xd): {
		(*k)++;
		break;
	}
	/* pitch bend event */
	case (0xe): {
		b[0] = 224 + midi_channel;
		b[1] = bytes[(*k)++];
		b[2] = bytes[(*k)++];

		insert_table(track, delta_time, b);
		break;
	}
	default: {
		break;
	}
	}
}

/*!
 * This function receives a MIDI file, reads its bytes, interprets them and
 * fills the table of MIDI events. It also assigns values to some important
 * global variables (like 'n_tracks' and 'ticks_per_beat'). It returns 0 in
 * case of success and returns 1 otherwise.
 */
int load_midi(FILE* input) {
	load_instruments();

	/* HEADER CHUNK */
	{
		int bit;
		unsigned char chunk_id[4];
		unsigned char chunk_size[4];
		unsigned char format_type[2];
		unsigned char number_of_tracks[2];
		unsigned char time_division[2];

		fread(chunk_id, 1, 4, input);
		fread(chunk_size, 1, 4, input);
		fread(format_type, 1, 2, input);
		fread(number_of_tracks, 1, 2, input);
		fread(time_division, 1, 2, input);

		if (bytes_to_decimal(format_type, 2) != 1) {
			fprintf(stderr, "error: mid file is not valid...\n");
			return 1;
		}

		n_tracks = bytes_to_decimal(number_of_tracks, 2);

		bit = (0x80 & time_division[0]) >> 7;

		time_division[0] = (time_division[0] & 0x7f);

		if (bit == 0) {
			ticks_per_beat = bytes_to_decimal(time_division, 2);
		} else if (bit == 1) {
			fprintf(stderr, "error: mid file is not valid...\n");
			return 1;
		}
	}

	/* TRACK CHUNK */
	{
		int i, size;
		unsigned char chunk_id[4];
		unsigned char chunk_size[4];
		unsigned char* track_event_data;

		int delta_time;
		int event;
		int k;
		short last;

		init_table();

		for (i = 0; i < n_tracks; i++) {
			fread(chunk_id, 1, 4, input);
			fread(chunk_size, 1, 4, input);

			size = bytes_to_decimal(chunk_size, 4);

			track_event_data = (unsigned char*) malloc(size
					* sizeof(unsigned char));

			fread(track_event_data, 1, size, input);

			delta_time = 0;
			k = 0;
			last = 0;

			while (k < size) {
				delta_time += variable_time(track_event_data, &k);

				if (track_event_data[k] >> 7) {
					last = event = track_event_data[k++];
				} else {
					event = last;
				}

				switch (event) {
				case (0xf0): {
					system_exclusive(track_event_data, &k);
					break;
				}
				case (0xff): {
					meta(track_event_data, &k, delta_time);
					break;
				}
				default: {
					midi(track_event_data, &k, event, (i + 1), delta_time);
					break;
				}
				}
			}

			free(track_event_data);
		}
	}

	if (fgetc(input) != EOF) {
		fprintf(stderr, "error: mid file is not valid...\n");
		return 1;
	}

	return 0;
}
